from pathlib import Path
import polars as pl
import duckdb
import sys
import re
from loguru import logger

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')


class PipelineText:
    def __init__(self, mode: str = ''):
        self.mode = mode

    @staticmethod
    def clean_text(data: pl.DataFrame, col: str = 'item_name') -> pl.DataFrame:
        regex = "[\(\[\<\"].*?[\)\]\>\"]"
        return data.with_columns(
            pl.col(col).map_elements(
                lambda x: re.sub(regex, "", x).lower().rstrip('.').strip(), return_dtype=pl.String
            )
            .alias(f'{col.lower()}_clean')
        )

    def run(self, data):
        # load data
        query = f"""select * from data"""
        df = duckdb.sql(query).pl()
        logger.info(f'[Data] Base Data {self.mode}: {df.shape}')

        df = (
            df
            .pipe(PipelineText.clean_text)
            .select(pl.all().name.prefix(f'{self.mode}_'))
            .drop_nulls()
        )
        logger.info(f'[Data] Join Images {self.mode}: {df.shape}')
        return df


def rm_all_folder(path: Path) -> None:
    for child in path.glob('*'):
        if child.is_file():
            child.unlink()
        else:
            rm_all_folder(child)
    path.rmdir()


def make_dir(folder_name: str | Path) -> None:
    """Make a directory if it doesn't exist"""
    if isinstance(folder_name, str):
        folder_name = Path(folder_name)
    if not folder_name.exists():
        folder_name.mkdir(parents=True, exist_ok=True)



from src.item_matching import Matching
from pathlib import Path
import polars as pl
from src.item_matching.build_index.func_img import PipelineImage


path = Path.home() / 'Downloads/item_match'

# db
path_db = path / 'db_0.parquet'
pipe = PipelineImage(path)
df_db = pl.read_parquet(path_db)
df_db, df_img_db = pipe.run(df_db, mode='db', download=False)
df_db.write_parquet(path / 'db_image_clean.parquet')

# q
df_q = df_db.clone()
df_q.columns = [f'q_{i.split('db_')[1]}' for i in df_db.columns]
df_q.write_parquet(path / 'q_image_clean.parquet')

json_stats = Matching(
    col_category='level1_kpi_category',
    path=path,
    path_query=path / 'q_image_clean.parquet',
    path_database=path / 'db_image_clean.parquet',
    query_batch_size=50_000,
    match_mode='image',
).run(export_type='csv')



from pathlib import Path
import polars as pl
import duckdb
import orjson
from tqdm import tqdm
import subprocess
import os
import sys
from loguru import logger
from .func import PipelineText, make_dir

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')


class PipelineImage:
    def __init__(
            self,
            path: Path,
            col_image: str = 'image_url',
            mode: str = ''
    ):
        self.path_image = path / 'download_img'
        self.col_image = col_image
        self.mode = mode

        # init path image
        make_dir(self.path_image)

    def download_images(self):
        folder = self.path_image / f'img_{self.mode}'
        if not folder.exists():
            os.chdir(str(self.path_image))
            command = (
                f"img2dataset --url_list={self.mode}_0.parquet "
                f"--output_folder=img_{self.mode}/ "
                f"--processes_count=16 "
                f"--thread_count=32 "
                f"--image_size=224 "
                f"--output_format=files "
                f"--input_format=parquet "
                f"--url_col={self.col_image} "
                f"--number_sample_per_shard=50000 "
            )
            subprocess.run(command, shell=True)

    def load_images(self) -> pl.DataFrame:
        # listing
        path = self.path_image / f'img_{self.mode}'
        lst_json = sorted(path.glob('*/*.json'))
        lst_file = [orjson.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]
        lst_img = [str(i) for i in tqdm(sorted(path.glob('*/*.jpg')), desc='Loading jpg in folder')]
        df = pl.DataFrame({
            f'{self.mode}_{self.col_image}': lst_file,
            f'{self.mode}_file_path': lst_img,
            f'{self.mode}_exists': [True] * len(lst_file),
        })

        logger.info(f'[Data] Load Images: {df.shape}')
        return df

    def run(
            self,
            data,
            download: bool = False,
            edit_img_url: bool = True,
    ):
        # edit url
        col_query = f",concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) {self.col_image}"
        if edit_img_url:
            col_query = f"images {self.col_image}"

        # load data
        query = f"""select *, {col_query} from data"""
        df = duckdb.sql(query).pl()
        logger.info(f'[Data] Base Data {self.mode}: {df.shape}')

        # download
        if download:
            df.write_parquet(self.path_image / f'{self.mode}_0.parquet')
            self.download_images()

        # load data image
        data_img = self.load_images()

        # join
        data = (
            df.drop(['images'])
            .pipe(PipelineText.clean_text)
            .select(pl.all().name.prefix(f'{self.mode}_'))
            .join(data_img, on=f'{self.mode}_{self.col_image}', how='left')
            .filter(pl.col(f'{self.mode}_exists'))
        )
        logger.info(f'[Data] Join Images {self.mode}: {data.shape}')
        return data, data_img



from tqdm import tqdm
from pathlib import Path
import json
import orjson
import gc

path = Path.home() / '/media/kevin/75b198db-809a-4bd2-a97c-e52daa6b3a2d/item_match/download_img/img_db'
lst_json = sorted(path.glob('*/*.json'))
gc.collect()
[json.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]
gc.collect()
[orjson.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]



from src.item_matching import Matching
from pathlib import Path


path = Path.home() / 'Downloads/item_match'
path_db = path / 'db_0.parquet'
path_q = path / 'db_0.parquet'
json_stats = Matching(
    col_category='level1_kpi_category',
    path=path,
    path_database=path_db,
    path_query=path_q,
    match_mode='text',
).run()



from pathlib import Path
import duckdb
import re
from time import perf_counter
from .build_index.func import rm_all_folder, make_dir
from .build_index.matching import BELargeScale


class Matching:
    def __init__(
            self,
            col_category: str,
            path: Path,
            path_database: Path = None,
            path_query: Path = None,
            query_batch_size: int = 500_000,
            match_mode: str = 'text_dense',
    ):
        self.path = path
        self.path_database = path_database
        self.path_query = path_query
        self.col_category = col_category
        self.query_batch_size = query_batch_size
        self.match_mode = match_mode

    def run(
            self,
            export_type: str = 'parquet',
            top_k: int = 10,
    ):
        """
        Run matching processes
        :param export_type: parquet | csv
        :param top_k: top k matches
        :return: json
        """
        # read query file to extract category
        query = f"""select distinct q_{self.col_category} as category from read_parquet('{self.path_query}')"""
        lst_category = duckdb.sql(query).pl()['category'].to_list()

        # Match
        be = BELargeScale(self.path, text_dense=True)
        if self.match_mode == 'image':
            be = BELargeScale(self.path, img_dim=True, query_batch_size=self.query_batch_size)

        path_match_result = self.path / 'result_match'
        make_dir(path_match_result)
        start = perf_counter()
        # Run
        for idx, cat in enumerate(lst_category):
            # read chunk cat
            query = f"""
            select * 
            from read_parquet('{self.path_database}') 
            where db_{self.col_category} = '{cat}'
            """
            chunk_db = duckdb.sql(query).pl()
            query = f"""
            select * 
            from read_parquet('{self.path_query}') 
            where q_{self.col_category} = '{cat}'
            """
            chunk_q = duckdb.sql(query).pl()
            print(f"🐋 Start matching by [{self.match_mode}] cat: {cat} {idx}/{len(lst_category)} - "
                  f"Database shape {chunk_db.shape}, Query shape {chunk_q.shape}")

            # check
            if chunk_q.shape[0] == 0 or chunk_db.shape[0] == 0:
                print(f'Database/Query have no data')
                continue

            cat = re.sub('/', '', cat)
            file_name = path_match_result / f'{cat}.{export_type}'
            if file_name.exists():
                print(f'File already exists: {file_name}')
                continue

            # match
            df_match = be.match(chunk_db, chunk_q, top_k=top_k)

            # export
            if export_type == 'parquet':
                df_match.write_parquet(file_name)
            else:
                df_match.write_csv(file_name)

            # remove caches
            rm_all_folder(self.path / 'index')
            rm_all_folder(self.path / 'result')
            for i in ['db', 'q']:
                rm_all_folder(self.path / f'{i}_array')
                rm_all_folder(self.path / f'{i}_ds')

        # update log
        time_perf = perf_counter() - start
        json_stats = {'time_perf': time_perf, 'path result': path_match_result}
        print(f'🐋 Your files are ready, please find here: {self.path}')
        return json_stats



import torch
from loguru import logger
import sys
from PIL import Image
from torch.nn import functional as F
from transformers import AutoProcessor, SiglipVisionModel
from FlagEmbedding import BGEM3FlagModel

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')
device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'


class Model:
    def __init__(self):
        self.device = device
        logger.info(f'[Model] Run on: {self.device}')

    def get_img_model(self):
        """Get Image Model"""
        model_id = 'google/siglip-base-patch16-224'
        img_processor = AutoProcessor.from_pretrained(model_id)
        img_model = SiglipVisionModel.from_pretrained(model_id).to(self.device)
        logger.info(f'Image model: {model_id}')
        return img_model, img_processor

    def get_text_model(self):
        model_id = 'BAAI/bge-m3'
        model = BGEM3FlagModel(model_id, use_fp16=True)
        logger.info(f'Text model: {model_id}')
        return model

    @staticmethod
    def pp_img(batch, model, processor, col: str) -> dict:
        images = [Image.open(i).convert('RGB') for i in batch[col]]
        inputs = processor(images=images, return_tensors='pt').to(device)
        with torch.inference_mode():
            outputs = model(**inputs)
        pooled_output = outputs.pooler_output
        embeddings = F.normalize(pooled_output, p=2, dim=1).cpu().numpy()
        return {'img_embed': embeddings}

    @staticmethod
    def pp_dense(batch, model, col: str) -> dict:
        embeddings = model.encode(
            batch[col],
            batch_size=512,
            max_length=80,
            return_dense=True,
            return_sparse=False,
            return_colbert_vecs=False
        )['dense_vecs']
        embeddings = F.normalize(torch.tensor(embeddings), p=2, dim=1).cpu().numpy()
        return {'dense_embed': embeddings}
